/***************************************************************************
* this is a class for pulling individual bits out of a data stream
* Most significant bits come first in the stream
***************************************************************************/
import java.io.*;

public class BitInputStream extends DataInputStream {

int buffer;
int nbits;

/***************************************************************************
*
***************************************************************************/
public BitInputStream(InputStream in) {

super(in);

nbits=0;

} // end of constructor

/***************************************************************************
* read the value of the next bit in the stream
***************************************************************************/
public int readBit() throws IOException {

/****************************************
* check if we need to refill the buffer *
****************************************/
if(nbits == 0) {
    buffer = read();
    if(buffer==-1) throw new EOFException();
    nbits=8;
}

/****************
* extract value *
****************/
int value = ( buffer & (1<<7) ) >> 7;

/****************************************
* take the read value out of the buffer *
****************************************/
buffer <<= 1;
--nbits;

return value;

} // end of readBit

/***************************************************************************
* read the next n bits from the stream into the least significant bits
* of an integer.
* n>32 is treated as n==32 and n<0 is treated as n=0
***************************************************************************/
public int readBits(int n) throws IOException {

int value=0;

for(int i=0; i<n; ++i) {

    value <<= 1;
    value |= readBit();
}

return value;

} // end of readBits method

/***************************************************************************
* read the next n bits from the stream into the least significant bits
* of an integer.
* n>32 is treated as n==32 and n<0 is treated as n=0
***************************************************************************/
public long readLongBits(int n) throws IOException {

long value=0;

for(int i=0; i<n; ++i) {

    value <<= 1;
    value |= readBit();
}

return value;

} // end of readLongBits method

/****************************************************************************
* read a number of unsigned bits and return them as a Number object.
* That object is either a Long or an Integer, depending on the number of
* bits requested.
* If more bits are requested than can be held in the positive part of a Long,
* the this methoid throws an IOException
****************************************************************************/
public Number readUnsignedBits(int n) throws IOException {

if(n>=64) throw new IOException("Can't read "+n+" bits at a time");

if(n<32) return new Integer(readBits(n));
else     return new Long(readLongBits(n));

} // end of readUnsignedBits method

/****************************************************************************
* read a number of unsigned bits and return them as a Number object.
* That object is either a Long or an Integer, depending on the number of
* bits requested.
* If more bits are requested than can be held in the positive part of a Long,
* the this methoid throws an IOException
****************************************************************************/
public Number readSignedBits(int n) throws IOException {

if(n>64) throw new IOException("Can't read "+n+" bits at a time");

if(n==64) return new Long(readLongBits(n));
if(n==32) return new Integer(readBits(n));

if(n>32) {
    /*******************
    * read into a long *
    *******************/

    if(readBitAsBoolean()) {
        /***********
        * negative *
        ***********/
        long value = Long.MIN_VALUE;
        value <<= (n-1);
        value |= readLongBits(n-1);
        return new Long(value);

    } else {
        /***********
        * positive *
        ***********/
        return new Long(readLongBits(n-1));
    }


} else {
    /*******************
    * read into an int *
    *******************/

    if(readBitAsBoolean()) {
        /***********
        * negative *
        ***********/
        int value = Integer.MIN_VALUE;
        value <<= (n-1);
        value |= readBits(n-1);
        return new Integer(value);

    } else {
        /***********
        * positive *
        ***********/
        return new Integer(readBits(n-1));
    }

} // end if bits will fit into an int

    
} // end of readSignedBits method





/***************************************************************************
* read a single bit and return true if it's value is 1
* or false otherwise
***************************************************************************/
public boolean readBitAsBoolean() throws IOException {

if(readBit() == 1) return true;
else               return false;

}

/****************************************************************************
* returns true if the bit buffer is empty - i.e. if the next read would
* come from the most significant bit of a byte.
***************************************************************************/
public boolean isByteAligned() { return nbits==0; }


/****************************************************************************
* returns the number of bits which must be read before isByteAligned 
* returns true
***************************************************************************/
public int getBitsToAlign() { return nbits; }


} // end of BitInputStream class
