import java.io.*;
import javax.swing.tree.*;

public abstract class DataStream extends DataBlock {

int parse_from;
int parse_to;

int nheader;
PlaceHolder begin;
PlaceHolder end;

protected DataNode node;

/****************************************************************************
* constructor
****************************************************************************/
public DataStream() { this(null);}
public DataStream(DataNode node) {

    this.node = node;

    parse_from=1;
    parse_to=-1;
} // end of constructor



/****************************************************************************
* set first and last parsed packet number
****************************************************************************/
public void setFirstParsedPacket(int first) { parse_from = first; }
public void  setLastParsedPacket(int  last) { parse_to   =  last; }

/****************************************************************************
* gives the file index of the first parsed packet
****************************************************************************/
public int getFirstParsed() {

    /****************************************************
    * return the set value if nothing has been read yet *
    ****************************************************/
    if(getChildCount()==0) return parse_from;

    /************************************************
    * return 1 if there is no begining place holder *
    ************************************************/
    if(begin==null || getIndex(begin) == -1 ) return 1;

    return begin.getPlaceCount() +1;

} // end of getFirstParsed method

/****************************************************************************
* gives the file index of the last parsed packet
****************************************************************************/
public int  getLastParsed() { 

    if(getChildCount()==0) return parse_to;


    /******************************************
    * make sure there is an end placeholder.
    * if not return "read to end" = -1
    *****************************************/
    if(end==null) return -1;
    int end_index = getIndex(end);
    if(end_index==-1) return -1;

    /**********************************************
    * find the index of the begin placeholder
    **********************************************/
    int begin_index = -1;
    if(begin!=null) begin_index = getIndex(begin);

    /***************************************************
    * if there is no begin placeholder, then fake one
    * taking the number of header packets into account
    ***************************************************/
    if(begin_index==-1) begin_index = nheader-1;

    int nparsed = end_index - begin_index -1;

    return getFirstParsed() + nparsed -1;

} // end of getLastParsed method


/***************************************************************************
* return the current number of parsed packets
***************************************************************************/
public int getParsedCount() {

    int first = getFirstParsed();
    int last  = getLastParsed();

    if(last==-1) {

        int nchildren = getChildCount();
        if(nchildren==0) return 0;

        return nchildren - treeIndex(first);
    } else {
        return last-first+1;
    }

} // end of getParsedCount method

/***************************************************************************
* return the number of unparsed packets at the end of the file
***************************************************************************/
public int getEndUnparsedCount() {

if(end==null || getIndex(end)==-1) return 0;
else                               return end.getPlaceCount();

} // end of getParsedCount method

/***************************************************************************
* return the number of unparsed packets at the end of the file
***************************************************************************/
public int getBeginningUnparsedCount() {

if(begin==null || getIndex(begin)==-1) return 0;
else                                   return begin.getPlaceCount();

} // end of getParsedCount method

/*****************************************************************************
* return the total number of parsed and unparsed packets
*****************************************************************************/
public int getCount() {

    return getBeginningUnparsedCount() + getParsedCount() + 
                 getEndUnparsedCount();

}



/*****************************************************************************
* translate a given packet file index into a tree index
*****************************************************************************/
private int treeIndex(int file_index) {

    int first_packet = nheader;

    int begin_index = -1;
    if(begin!=null) begin_index=getIndex(begin);

    int first_index;
    if(begin_index!=-1) first_index=begin_index+1;
    else                first_index = nheader;

    return first_index +file_index - getFirstParsed();


} // end of treeIndex method


/****************************************************************************
* method for reading the packets in a stream
****************************************************************************/
public void read(BitInputStream in) throws IOException {

/*****************************
* clear out any old contents *
*****************************/
removeAllChildren();
begin=null;
end=null;


try {

    /********************************
    * read a header if there is one *
    ********************************/
    readHeader(in);
    nheader = getChildCount();

    /**************************************************
    * read a placeholder at the beginning of the file *
    **************************************************/
    if(parse_from>1) {
        begin = new PlaceHolder(node, parse_from-1);
        try { begin.read(in); }
        catch(EOFException e) {}
        add(begin);
    }

    /**********************
    * read parsed packets *
    **********************/
    for(int i=parse_from; parse_to<0 || i<=parse_to; ++i) {

        /*********************************************
        * read a data block and add it to the stream *
        *********************************************/
        add(readUnit(in));
 
    }

    /********************************
    * read a placeholder at the end *
    ********************************/
    if(parse_to>0) {
        end = new PlaceHolder(node);
        try { end.read(in); }
        catch(EOFException e) {}
        add(end);
    }

        

} catch(EOFException e) {} // this is normal
  catch(OutOfMemoryError e) {

    removeAllChildren();
    begin=null;
    end=null;
    throw new IOException("Ran out of memory");
}

} // end of read method

/****************************************************************************
* changed the parsed window to the given set of file indices. Note
* the first packet in the file has index 1. Both to and from are inclusive.
****************************************************************************/
public void rewindow(int from, int to) {

    /****************************************
    * figure out how many packets are where *
    ****************************************/
    int last_begin = getBeginningUnparsedCount();
    int nparsed = getParsedCount();
    int last_parsed = last_begin+nparsed;
    int last_packet = last_parsed + getEndUnparsedCount();


    /**********************************
    * make sure the indices are valid *
    **********************************/
    if(from<1) from=1;
    if(from>last_packet) from=last_packet;
    if(to<from) to=from;
    if(to>last_packet) to=last_packet;

    if(to<=last_begin ) {
        /********************************************
        * new interval is entirely within beginning *
        ********************************************/
        moveToEnd(nparsed);
        moveFromBeginningToEnd(last_begin-to);
        moveFromBeginning(to-from+1);

    } else if(to<=last_parsed ) {
        if(from<=last_begin) {
            /********************************
            * straddling the begin boundary *
            ********************************/
            moveToEnd(last_parsed - to);
System.out.println("moving from beginning "+(from - last_begin+1));
            moveFromBeginning(last_begin-from+1);
        } else {
            /**********************************
            * entirely within old parsed area *
            **********************************/
            moveToEnd(last_parsed-to);
            moveToBeginning(from-last_begin-1);
        }

    } else {
        if(from<=last_begin) {
            /********************************************
            * new region completely contains old region *
            ********************************************/
            moveFromBeginning(last_begin-from+1);
            moveFromEnd(to-last_parsed);

        } else if(from<=last_parsed) {
            /******************************
            * straddling the end boundary *
            ******************************/
            moveToBeginning(from-last_begin-1);
            moveFromEnd(to-last_parsed);
        } else {
            /*********************************
            * entirely within the end region *
            *********************************/
            moveToBeginning(nparsed);
            moveFromEndToBeginning(from-last_parsed-1);
            moveFromEnd(to-from+1);
        }
    }

} // end of rewindow method
        



/****************************************************************************
* remove a number of packets from the tree and write them to a new placeholder
* "index" is the tree index of the first packet to unload
****************************************************************************/
private PlaceHolder unloadToPlaceHolder(int index, int count) {

//System.out.println("unloading "+count +" from index "+index);

try {

    /******************************************************
    * take the packets out of the tree and write them to 
    * a byte array
    ******************************************************/
    ByteArrayOutputStream bytes = new ByteArrayOutputStream();
    BitOutputStream out = new BitOutputStream(bytes);

    for(int i=0; i<count; ++i) {

        get(index).write(out);
        remove(index);
    }


    /***************************************************************
    * make a new place holder and read the data in the byte array 
    * into it
    ***************************************************************/
    PlaceHolder holder = new PlaceHolder(node, count);

    BitInputStream in = new BitInputStream(
                        new ByteArrayInputStream(bytes.toByteArray() ));

    holder.read(in);

    return holder;

} catch(IOException e) {
    System.err.println(e);
    return null; 
}

} // end of unloadToPlaceHolder method

    

/****************************************************************************
* read a number of packets from a placeholder into the tree. The first 
* packet will be placed at the given tree index. ****************************************************************************/
private void loadFromPlaceHolder(PlaceHolder holder, int index) {

//System.out.println("loading from placeholder");

try {

    /**********************************************************
    * make a stream to read from the placeholder's byte array *
    **********************************************************/
    BitInputStream in = new BitInputStream(
                        new ByteArrayInputStream(holder.getByteArray()));

    /*******************
    * read the packets *
    *******************/
    int count = holder.getPlaceCount();
    System.out.println("    loading count="+count);
    for(int i=0; i< count; ++i) {
        DataNode node = readUnit(in);

        insert(node, index++);
    }

} catch(IOException e) {
    System.err.println(e);
}


} // end of loadFromPlaceHolder method


/****************************************************************************
* take a given number of parsed packets and move them to the beginning 
* placeholder
****************************************************************************/
private void moveToBeginning(int count) {

//System.out.println("moving "+count+" to beginning");

/*************************************************
* do nothing if we're not asked to move anything *
*************************************************/
if(count==0) return;

/************************************
* do nothing if we have no contents *
************************************/
if(getChildCount()==0) return;

/*************************************************
* find the beginning placeholder if there is one *
*************************************************/
int begin_index=-1;
if(begin!=null) begin_index = getIndex(begin);

/*************************************
* find the index of the first packet *
*************************************/
int first_index;
if(begin_index==-1) first_index=nheader;
else                first_index=begin_index+1;

/****************************************************
* remove the packets and move them to a placeholder *
****************************************************/
PlaceHolder holder = unloadToPlaceHolder(first_index, count);

/***********************************************************
* merge the new placeholder with the beginning placeholder *
***********************************************************/
if(begin_index==-1) insert(begin=holder,nheader);
else                begin.append(holder);


} // end of moveToBeginning method


/****************************************************************************
* take a given number of units from the ending place holder and parse them
****************************************************************************/
private void moveFromBeginning(int count) {

    if(count==0) return;

    int begin_index = -1;
    if(begin!= null) begin_index = getIndex(begin);
    if(begin_index==-1) return;

    int insert_index = begin_index+1;

    PlaceHolder holder = begin.split(begin.getPlaceCount() - count);
    loadFromPlaceHolder(holder, insert_index);

    if(begin.getPlaceCount()==0) remove(begin);

} // end of moveFromEnd method


/****************************************************************************
* move a given number of units from the endig place holder to the
* beginning placeholder. This is only appropriate if there are no parsed
* packets
****************************************************************************/
private void moveFromEndToBeginning(int count) {

if(count==0) return;

    /********************************************
    * do nothing if there's no end place holder *
    ********************************************/
    if(end== null || getIndex(end)==-1) return;

    /*******************************
    * take the data out of the end *
    *******************************/
System.out.println("splitting end");
    PlaceHolder holder = end.split(count);
System.out.println("done splitting end");

    /**********************************
    * put the data into the beginning *
    **********************************/
    if(begin==null || getIndex(begin)==-1) insert(begin=end, nheader);
    else                                   begin.append(end);

    /**********************************
    * put in the new end place holder *
    **********************************/
    int end_index = getIndex(end);
    remove(end);

    if(holder.getPlaceCount()>0) insert(end=holder, end_index);
    else                         end=null;


} // end of moveFromEndToBeginning method

/****************************************************************************
*
****************************************************************************/
private void moveFromBeginningToEnd(int count) {

if(count==0) return;

if(begin==null || getIndex(begin)==-1) return;

System.out.println("    before splitting begin="+begin);
PlaceHolder holder = begin.split(begin.getPlaceCount()-count);
System.out.println("    after splitting begin="+begin+" holder="+holder);

if(end==null || getIndex(end)==-1) insert(end=holder, getChildCount());
else {
    holder.append(end);
    remove(end);
    insert(end=holder, getChildCount());
}



} // end of moveFromEndToBeginning method

/****************************************************************************
* move the given number of parsed packets to the end place holder
****************************************************************************/
private void moveToEnd(int count) {

if(count==0) return;

/************************************
* do nothing if we have no contents *
************************************/
if(getChildCount()==0) return;

/*************************************************
* find the end placeholder if there is one *
*************************************************/
int end_index=-1;
if(end!=null) end_index = getIndex(end);

/*************************************
* find the index of the first packet *
*************************************/
int first_index;
if(end_index==-1) first_index=getChildCount()-1;
else              first_index=end_index-1;

first_index -= count-1;

/****************************************************
* remove the packets and move them to a placeholder *
****************************************************/
PlaceHolder holder = unloadToPlaceHolder(first_index, count);

/*****************************************************
* merge the new placeholder with the end placeholder *
*****************************************************/
if(end_index==-1) insert(end=holder,getChildCount());
else {
    holder.append(end);
    remove(end);
    add(end=holder);
}


} // end of moveToEnd method



/****************************************************************************
* take a given number of units from the ending place holder and parse them
****************************************************************************/
private void moveFromEnd(int count) {

//System.out.println("moving "+count+" from end");

int end_index = -1;
if(end!= null) end_index = getIndex(end);
if(end_index==-1) return;

int insert_index = end_index;

PlaceHolder holder = end.split(count);
//System.out.println("top part: "+end);
loadFromPlaceHolder(end, insert_index);

end_index = getIndex(end);
remove(end);
if(holder.getPlaceCount()>0) insert(end=holder, end_index);
else                         end=null;

} // end of moveFromEnd method


/****************************************************************************
* string representation for display
****************************************************************************/
public String toString() { return getClass().getName(); }

/****************************************************************************
* read a header at the beginning of the file
****************************************************************************/
protected void readHeader(BitInputStream in) throws IOException {}

/****************************************************************************
* read a single "unit" of data from the stream
* subclasses usually override this method, which only throws an EOFException
****************************************************************************/
protected DataNode readUnit(BitInputStream in) throws IOException {

    throw new EOFException();

} // end of readUnit method

/****************************************************************************
*
****************************************************************************/
public DataNode find(String pattern, boolean just_window, 
                     DataNode start, int depth,
                     ProgressListener progress) {

/****************************************
* determine the total number of packets *
****************************************/
int total;
if(just_window) total = getChildCount();
else {
    total = nheader + getBeginningUnparsedCount() + getParsedCount()+ 
                             getEndUnparsedCount();
}

/********************
* loop over packets *
********************/
int index;
if(start==null) index=0;
else index=getIndex(start)+1;
for(; index<getChildCount(); ++index) {
    DataNode node = get(index);

    if(node==begin || node == end) {
        /************************
        * this is a placeholder *
        ************************/
        if(just_window) continue;

        int offset=index;
        if(node==end) offset+= getBeginningUnparsedCount();

        int found = searchPlaceHolder((PlaceHolder)node, pattern, depth, 
                                      progress, offset, total);
        if(found >=0) {
            /*******************************************
            * found it - rewindow and then do a quick
            * search to find it within the window
            *******************************************/
            int from = offset-nheader+found -1;
            rewindow(from, from+getParsedCount() );
            return find(pattern, true, get(0), depth, null);
        } else if(found==-2) {
            /*********
            * cancel *
            *********/
            return null;
        }
    }

    /******************
    * update progress *
    ******************/
    if(progress != null) {
        if(progress.cancel()) return null;
        progress.updateProgress((100*index)/total);
    }

    /****************************
    * search the current packet *
    ****************************/
    DataNode found = node.find(pattern, depth);
    if(found!= null) return found;


} // end of loop over contents of stream

return null;

} // end of find method


/*****************************************************************************
* returns -1 if nothing was found, -2 if the search was canceled
*****************************************************************************/
private int searchPlaceHolder(PlaceHolder placeholder, String pattern, 
                                   int depth, ProgressListener progress,
                                   int offset, int total) {

    BitInputStream in = new BitInputStream(
                        new ByteArrayInputStream(placeholder.getByteArray()));

    try {
        for(int i=0; true; ++i) {

            if(progress!=null) {
                if(progress.cancel() ) return -2;
                progress.updateProgress((100*(i+offset))/total);
            }


            DataNode node = readUnit(in);
            DataNode found = node.find(pattern, depth);
            if(found != null) return i;
        }
    } catch(IOException e) {}

    return -1;

} 

} // end of DataStream class
