import java.io.*;

/***************************************************************************
* This is a class for reading one data stream and translating it into 
* another format. This is an abstract class, since subclasses muyst defint
* the actual decoding method.
**************************************************************************/
public abstract class Filter extends PipedInputStream implements Runnable {

private BitInputStream in;
private boolean running;

/****************************************************************************
* Create a new object and start the extracting thread
****************************************************************************/
public Filter(BitInputStream in) {

    this.in = in;
    running=false;


} // end of constructor


/****************************************************************************
* This method must be called at the end of each sub-class's constructor.
* It starts the decoding thread.
****************************************************************************/
protected void start() {

    if(running) return;

    new Thread(this).start();
    waitUntilPipeReady();

}

/****************************************************************************
*
****************************************************************************/
synchronized private  void signalPipeReady() { notifyAll(); }

/****************************************************************************
*
****************************************************************************/
synchronized private  void waitUntilPipeReady() {

    try { wait(); } 
    catch(InterruptedException e) {}
}

/****************************************************************************
* Subclasses must define this method to read some part of the input stream
* and write the decoded data to the output stream.
* This method will be called repeatedly until the data have been decoded.
****************************************************************************/
abstract protected void decode(BitInputStream in, OutputStream out) throws IOException;

/****************************************************************************
* This method implements the Runnable interface. This is where all the
* decoding is done.
****************************************************************************/
public void run() {

/*************************************************
* make sure only one thread is running at a time *
*************************************************/
if(running) return;
running=true;

/*********************************************************
* define the output pipe here, so that we can close it
* if we get an exception 
*********************************************************/
PipedOutputStream out=null;

try {
    /****************************************
    * create a pipe connected to this class *
    ****************************************/
    out = new PipedOutputStream(this);
    signalPipeReady();

    /****************************************
    * read packets until we hit end of file *
    ****************************************/
    while(true) { decode(in,out); }

} catch(EOFException e) {
    /*****************************************************
    * close the pipe on an EOF
    * if the close fails, just kill the thread,
    * which will probably throw an IOException on the 
    * reading side
    *****************************************************/
    if(out==null) return;

    try { out.close(); }
    catch(IOException e2) { return; }

} catch(IOException e) {
    /**************************************************************
    * on error, flush the output stream and return to kill the
    * thread. That will cause an IOException to be thrown
    * in the reader's thread.
    *************************************************************/
    if(out==null) return;

    try { out.flush(); }
    catch(IOException e2) {}
    return;
}


} // end of run method


} // end of Filter class
