import java.io.*;

/****************************************************************************
* Used to mark a chunk of data which has not actually been read.
******************************************************************************/
public class PlaceHolder extends RawDataBlock {

DataNode node;

int npackets;
int[] offsets;


/***************************************************************************
* constructor
***************************************************************************/
PlaceHolder(DataNode node) { this(node, -1); }

PlaceHolder(DataNode node, int npackets) { 
    super(0);
    this.npackets = npackets;
    this.node=node;

    if(npackets>0) offsets = new int[npackets];
    else           offsets = new int[100];
}


/*****************************************************************************
* set an offset index, resizing the offsets array if necessary
*****************************************************************************/
private void setOffset(int index, int offset) {

//System.out.println("set offset index="+index+
//                   " offset="+offset+
//                   " offsets.length="+offsets.length);

    if(index>=offsets.length) {
        /**********************************
        * have to resize the offset array *
        **********************************/
//System.out.println("    resizing offsets");
        int[] new_offsets = new int[offsets.length*2];
        System.arraycopy(offsets, 0, new_offsets, 0, offsets.length);
        offsets=new_offsets;
    }

    offsets[index] = offset;
}

/***************************************************************************
* Return the number of places being held
***************************************************************************/
public int getPlaceCount() { return npackets; }


/***************************************************************************
* read the placeholder - actually we just scan through the data until
* we either get to the desired number of packets or we hit the end of file.
***************************************************************************/
public void read(BitInputStream in) throws IOException {


ByteArrayOutputStream data = new ByteArrayOutputStream();
BitOutputStream out = new BitOutputStream(data);

int size=0;
int i=0;
try {

    for(i=0; npackets<0 || i<npackets; ++i) {
            node.read(in);
            setOffset(i+1, offsets[i] + node.size()/8 );
            node.write(out);
    }

    this.data = data.toByteArray();
    this.size = this.data.length;

} catch(IOException e) {
    npackets=i;
    this.data = data.toByteArray();
    this.size = this.data.length;
    throw e;
}

} // end of read method

/*****************************************************************************
* returns the byte offset in the data array of the start of the given
* packet number
*****************************************************************************/
private int find(int packet) { return offsets[packet]; }

/*****************************************************************************
* split this placeholder in two at the given offset in *places*
* this placeholder will be shortened to represent the beginning
* and a new placeholder will be returned to represent the end.
* The argument gives the number of packets which will remain in this
* placeholder.
*****************************************************************************/
public PlaceHolder split(int packets) {

//System.out.println("splitting "+packets+" from "+this);

    /******************************************
    * make sure the packet split is in bounds *
    ******************************************/
    if(npackets<0) return null;
    if(packets<0) packets=0;
    if(packets>npackets) packets=npackets;

    /***************
    * special case *
    ***************/
    if(packets==npackets) return new PlaceHolder(node, 0);

    /************************************************************
    * create a new placeholder to represent the data at the end *
    ************************************************************/
    PlaceHolder holder = new PlaceHolder(node, npackets-packets);

    /*****************************************
    * copy the data into the new placeholder *
    *****************************************/
    int offset = find(packets);
    int length = data.length - offset;

    holder.resize(length);
    System.arraycopy(data,offset, holder.data,0, length);

    /************************************************
    * now copy the offsets into the new placeholder *
    ************************************************/
    for(int i=packets; i<npackets; ++i) {
//System.out.println("    i="+i);
        holder.setOffset(i-packets, find(i)-offset );
    }

    /***********************************************************************
    * resize this place holder to represent only the data at the beginning *
    ***********************************************************************/
    resize(offset);
    npackets=packets;

    return holder;


} // end of split method


/*****************************************************************************
* append a placeholder onto the end of this one
*****************************************************************************/
public void append(PlaceHolder holder ) {

    int old_size = size()/8;
    int new_size = (size() + holder.size())/8;
    resize(new_size);

    /****************
    * copy the data *
    ****************/
    System.arraycopy(holder.data,0, data,old_size, new_size-old_size);

    /**************************
    * adjust the packet count *
    **************************/
    int old_npackets = npackets;
    npackets += holder.getPlaceCount();


    /*******************
    * copy the offsets *
    *******************/
    for(int i=old_npackets; i<npackets; ++i) {
        setOffset(i, old_size + holder.find(i-old_npackets) );
    }

} // end of append method



/*****************************************************************************
* string representation
*****************************************************************************/
public String toString() { 

    return npackets + " unparsed "+
           node.getClass().getName()+
           " ("+size()+" bytes)";
}

/*****************************************************************************
*
*****************************************************************************/
//public BitInputStream getBeginningStream() {
//System.out.println("getting beginning stream");
//    return new BitInputStream(new BeginningStream(this) );
//}

/*****************************************************************************
******************************************************************************
* A data stream for reading from the beginning of the placeholder.
* The data read will be removed from the placeholder when the 
* stream is closed.
*****************************************************************************/
public class BeginningStream extends ByteArrayInputStream {

public BeginningStream(PlaceHolder holder) {
    super(holder.data);
}

public void close() throws IOException {

    System.out.println("read "+pos+" bytes from beginning");

    resize(size-pos, pos);
    super.close();
}

} // end of BeginningStream inner class
    
/*****************************************************************************
* for testing
*****************************************************************************/
public static void main(String[] args) {
}





} // end of PlaceHolder class
