import java.io.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;

class RawDataBlock extends DataNode {

/*******
* data *
*******/
protected byte[] data;
protected int size;

protected int preferred_size;

private DataNode display;



/*****************************************************************************
* constructor
*****************************************************************************/
public RawDataBlock( int nbytes ) {

    this(new byte[nbytes]);

} // end of constructor

/*****************************************************************************
* constructor. If set_preferred=true, sets the preferred size to the
* current size.
*****************************************************************************/
public RawDataBlock(int nbytes, boolean set_preferred) {

    this(nbytes);
    if(set_preferred) setPreferredSize(nbytes);
}


/*****************************************************************************
* constructor from an existing array - just uses that array by reference
*****************************************************************************/
public RawDataBlock( byte[] data ) {

this.data = data;
size = data.length;

display = makeDisplay();
insert(display, 0 );

preferred_size=-1;


} // end of constructor


/*****************************************************************************
* 
*****************************************************************************/
protected DataNode makeDisplay() { return new DataBlockDisplay(); }


/*****************************************************************************
* set the number of bytes that will always be attempted to be read
* regardless of the current size of the data block
* If the preferred size is -1 (the default), then the current
* size will always be read
*****************************************************************************/
public void setPreferredSize(int nbytes) { preferred_size = nbytes; }


/*****************************************************************************
* 
*****************************************************************************/
public byte[] getByteArray() { return data; }


/*****************************************************************************
* change the size of the data array (old data are saved)
*****************************************************************************/
public void old_resize(int new_size) {

if(new_size<0) new_size=0;
if(new_size == size) return;

byte[] old_data = data;

data = new byte[new_size];


int min_size = size;
if(min_size > new_size) min_size = new_size;

for(int i=0; i<min_size; ++i) {
    data[i] = old_data[i];
}

size = new_size;

} // end of resize method

/*****************************************************************************
* Change the size of the array, adding or removing all bytes from the end
*****************************************************************************/
public void resize(int new_size) { resize(new_size,0); }

/*****************************************************************************
* Change the size of the data array. Old data are saved. The offset
* argument gives the number of bytes which will be added to or removed
* from the beginning of the array. 
*****************************************************************************/
public void resize(int new_size, int offset) {

if(new_size<0) new_size=0;
if(new_size == size) return;

int max_offset = Math.abs(size-new_size);
if(offset>max_offset) offset = max_offset;

/*******************
* make a new array *
*******************/
byte[] old_data = data;
data = new byte[new_size];

/********************************
* copy the data if there is any *
********************************/
if(size>0) {
    if(new_size<size) {
        /******************
        * getting smaller *
        ******************/
        System.arraycopy(old_data,offset, data,0, new_size);
//        for(int i=0; i< new_size; ++i) {
//            data[i] = old_data[i+offset];
//        }
    } else {
        /*****************
        * getting bigger *
        *****************/
        System.arraycopy(old_data,0, data,offset, size);
//        for(int i=0; i<size; ++i) {
//            data[i+offset] = old_data[i];
//        }
    }
}

/**************************
* adjust some bookkeeping *
**************************/
size = new_size;

} // end of resize method




/*****************************************************************************
* read the data
*****************************************************************************/
public void read(BitInputStream   in) throws IOException {

    /**************************************************
    * resize to the preferred size if it has been set *
    **************************************************/
    if(preferred_size>=0) resize(preferred_size); 

    int nread = in.read(data,0,size);
    if(nread==-1) throw new EOFException();


    /********************************************************
    * we didn't get all the bytes, so wait a bit
    * and try it again. We keep this up until we get EOF
    * or we get all the bytes 
    ********************************************************/
    while(nread<size) {

        Thread.yield();

        int n = in.read(data,nread,size-nread);
//System.out.println("read "+n+" this time was nread="+nread+" size="+size);
        if(n==-1) {
            /************************************************
            * we hit the end of the file, but we were
            * able to read some bytes above, so don't
            * throw an EOFException. We do want to resize 
            * the array though
            ************************************************/
            resize(nread);
            return;
        } else {
            /***********************************************
            * we got some more bytes. See if we got enough *
            ***********************************************/
            nread += n;
 
        }

    } // end of loop blocking until we read everything
        


} // end of read method


/*****************************************************************************
* write the data
*****************************************************************************/
public void write(BitOutputStream out) throws IOException {

    out.write(data,0,size);

} // end of write method

/*****************************************************************************
* return the number of *bits* in the array
*****************************************************************************/
public int size() { return 8*size; }

/*****************************************************************************
* string representation
*****************************************************************************/
public String toString() { 
    return size+" bytes of raw data";
}

/****************************************************************************
*****************************************************************************
* embedded class for displaying data
****************************************************************************/
class DataBlockDisplay extends DataNode implements DataNodeEditor, KeyListener {

    /*******
    * data *
    *******/
    static final int columns = 32;
    JTextArea text;
    JTextArea editor;
    java.util.List listeners;

    /************************************************************
    * constructor 
    ************************************************************/
    DataBlockDisplay() {

        listeners = new ArrayList();

    } // end of constructor

    /**************************************************************
    * return component used for drawing
    **************************************************************/
    public Component displayComponent(boolean selected, boolean expanded, 
                                  int row, boolean hasFocus) {

        if(text == null) {
            /************************************
            * we need to create the text area
            ************************************/
            int rows = (size+columns-1)/columns;
       
            text = new JTextArea(rows, columns*3+1);
            text.setFont(new Font("Courier", Font.PLAIN, 12) );

            for(int i=0; i<size; ) {

                for(int col=0; col<columns && i<size; ++col) {

                    /*********************************************
                    * get the two digit hex string for this byte *
                    *********************************************/
                    String hex = Integer.toHexString(data[i++]);
                    int length = hex.length();
                    if(length < 2) hex = "0"+hex;
                    if(length > 2) hex = hex.substring(length-2);

                    /**********************************
                    * add the string to the text area *
                    **********************************/
                    text.append(hex+" ");
                }

                text.append("\n");

            }
        } // end if we need to create the text area

        return text;

    } // end of displayComponent method

    /***************************************************************************
    * editing methods 
    ***************************************************************************/
    public DataNodeEditor getEditor(int row) { return this; }

    /***************************************************************************
    * allow a new value to be set  - called by tree after editing
    ***************************************************************************/
    public void setUserObject(Object o) {

        /**********************************
        * force the display to get remade *
        **********************************/
        text=null;

        /*****************************************
        * loop over all the tokens in the string *
        *****************************************/
        StringTokenizer tokenizer = new StringTokenizer(o.toString());

        /**************************************
        * resize the data array if we need to *
        **************************************/
        resize(tokenizer.countTokens() );

        for(int i=0; tokenizer.hasMoreTokens(); ++i ) {

            String token = tokenizer.nextToken();

            try {
                data[i] = (byte)Integer.parseInt(token,16);
            } catch(NumberFormatException e) {}

        } // end of loop over tokens


    } // end of setUserObject method


    /*********************************************************************
    * the following methods implement the DataNodeEditor interface
    *********************************************************************/
    public Object getEditorValue() { return editor.getText(); }
    public Component getEditorComponent() { 

        editor = new JTextArea(text.getText());
        editor.setFont(new Font("Courier", Font.PLAIN, 12) );
        editor.setBorder(BorderFactory.createLineBorder(Color.black));
        editor.addKeyListener(this);
        return editor;

    }

    public void    addActionListener(ActionListener l) {listeners.add(l);}
    public void removeActionListener(ActionListener l) {listeners.remove(l);}

    /*****************************************************************
    * the following methods implement KeyListener so that we can
    * listen to the editor text area to intercept  returns
    *****************************************************************/
    public void keyReleased(KeyEvent e) {}
    public void keyTyped(KeyEvent e) {}
    public void keyPressed(KeyEvent e) {

        if(e.getKeyChar() == '\n' ) {
            /***********************************************************
            * hitting return means we're done editing 
            * so fire off an action event to anyone listening
            * note we have to make a copy of the listener list first 
            * so that the listeners may remove themselves while
            * we are looping
            **********************************************************/
            java.util.List tmp = new ArrayList(listeners);
            for(Iterator it = tmp.iterator(); it.hasNext(); ) {
                ActionListener l = (ActionListener)it.next();
                l.actionPerformed(new ActionEvent(this,0,"Done Editing") );
            } // end of loop over listeners
        } // end if return was pressed

    } // end of keyPressed method


    /************************************************************
    * the following methods must be defined for a DataNode
    * we define them to represent an empty block of data
    * Since this is just a display which doesn't actually hold
    * any data
    ************************************************************/
    public void  read(BitInputStream   in) throws IOException {}
    public void write(BitOutputStream out) throws IOException {}

    public int size() { return 0; }

} // end of DataBlockDisplay embedded class


} // end of RawDataBlock
