import java.io.*;
import java.util.*;

/****************************************************************************
* This class helps locate information about the environment in which
* xing is being run.
* It has a main method which provides an interface for this information
* for a shell script
****************************************************************************/
public class XingIniter {

static final String formats_subdir = "formats";
static final String parent_filename = "parent";
static final String help_filename = "help";

File instalation_dir;
File local_config_dir;
Properties defaults;

FileSearchPath format_path;

/**************************************************************************
*
**************************************************************************/
public XingIniter() {

/*********************************
* find the instalation directory *
*********************************/
File me = FileSearchPath.getClassPath().find(getClass().getName()+".class");
if(me==null) { 
    System.err.println("Couldn't find instalation directory");
    System.exit(1);
}

instalation_dir = me.getParentFile();

/******************************
* load the default properties *
******************************/
defaults = new Properties();

try {
    FileInputStream in = new FileInputStream(new File(instalation_dir, 
                                                      "config")       );

    defaults.load(in);
} catch(IOException e) {
    System.err.println("could not read default properties");
    System.err.println(e.getMessage() );
    System.exit(1);
}


/*****************************************
* find the local configuration directory *
*****************************************/
local_config_dir = new File( new File(System.getProperty("user.home")),
                             defaults.getProperty("local.dir")         );


/****************************************
* search path for format specifications *
****************************************/
format_path = new FileSearchPath();

if(local_config_dir.isDirectory() ) {
    format_path.add(new File(local_config_dir, formats_subdir) );
}

format_path.add(new File(instalation_dir, formats_subdir) );

} // end of constructor

/**************************************************************************
*
**************************************************************************/
public File getInstalationDirectory() { return instalation_dir; }

/**************************************************************************
*
**************************************************************************/
public File getLocalConfigDirectory() { return local_config_dir;} 

/**************************************************************************
*
**************************************************************************/
public FileSearchPath getFormatPath() { return format_path; }


/**************************************************************************
* return the name of the parent of a given format
* returns null if there is no parent.
* Throws an IOException if there was trouble reading the name of the
* parent
**************************************************************************/
public String formatParent(File dir) throws IOException{

/*****************************************************************
* see if there is a parent for this format
*****************************************************************/
String line = null;
File parent = new File(dir,parent_filename);
if(parent.isFile() ) {
    /***********************************************************
    * there is a parent, so read its name from the parent file
    ***********************************************************/
    BufferedReader reader = new BufferedReader(new FileReader(parent));
    line = reader.readLine();
    reader.close();
}

return line;

} // end of formatParent method


/**************************************************************************
* recursive function which returns the classpath for a given format name
* and then recursively appends the classpath for the parent format
**************************************************************************/
public String makeClassPath(String format) throws IOException {


/*******************************************************************
* if no format is specified, just return the instalation directory *
*******************************************************************/
if(format==null) return instalation_dir.toString();


/******************************************
* find the directory for the named format *
******************************************/
File dir = format_path.findDirectory(format);
if(dir==null) throw new IOException("Unknown format "+format);

String classpath = dir.toString();

/******************************************************
* get the system-dependant system separator character *
******************************************************/
String separator = System.getProperty("path.separator");

/******************************************
* recursively add the path for the parent *
******************************************/
return classpath+separator+makeClassPath(formatParent(dir));

    
} // end of makeClassPath method

/**************************************************************************
* return an array of known format names
**************************************************************************/
public String[] listFormats() {

/********************************************
* list everything in the format directories *
********************************************/
SortedSet set = format_path.list(null);


/***************************
* make an array of strings *
***************************/
String[] formats = new String[set.size()];
int i=0;
for(Iterator it = set.iterator(); it.hasNext(); ) {
    String format = (String)it.next();

    formats[i++] = format;

}

return formats;


} // end of listFormats method



/**************************************************************************
* Return the classpath for a given format named on the command line
**************************************************************************/
public static void main(String[] args) {

XingIniter init = new XingIniter();

/***********************************************************
* the first argument names the information you want to get *
***********************************************************/
if(args.length==0) {
    System.exit(1);
}

if(args[0].equals("classpath") ) {
    /****************************************************
    * print the CLASSPATH for the format named in the
    * second argument
    ****************************************************/
    String format=null;
    if(args.length>=2) format = args[1];

    try {
        System.out.println(init.makeClassPath(format) );
    } catch(IOException e) {
        System.err.println("Error Determining CLASSPATH");
        System.err.println(e);
        System.exit(2);
    }

} else if(args[0].equals("list") ) {
    /***********************************
    * list all known formats
    ***********************************/
    String[] formats = init.listFormats();
    for(int i=0; i<formats.length; ++i) {
        System.out.println(formats[i]);
    }
} else if(args[0].equals("help") ){
    /******************************************************************
    * print information about the format named in the second argument *
    ******************************************************************/
    String format = args[1];
    System.out.println("Format: "+format);

    /********************************************
    * find the directory containing this format *
    ********************************************/
    File dir = init.getFormatPath().findDirectory(format);
    if(dir==null) {
        System.out.println("Unknown format");
        System.exit(0);
    }

    /****************************************
    * print the name of the format's parent *
    ****************************************/
    try {
        String parent = init.formatParent(dir);
        if(parent==null) parent = "none";
        System.out.println("Parent: "+parent);
    } catch(IOException e) {
        System.out.println("Parent: error determining name");
    }

    /**************
    * skip a line *
    **************/
    System.out.println();

    /**************************************
    * print the help file for this format *
    **************************************/
    File help = new File(dir, help_filename);
    if(! help.isFile() ) {
        System.out.println("No additional information available");
    } else {
        try {
            BufferedReader reader = new BufferedReader(new FileReader(help));
            String line;
            while((line=reader.readLine()) != null ) {
                System.out.println(line);
            }
        } catch(IOException e) {
            System.out.println("Error reading hep information");
        }
    }


}


System.exit(0);

} // end of main method


} // end of XingIniter class
